from typing import Any, Callable, Self, overload

import numpy as np
import typeguard
from numpy.typing import DTypeLike

from . import _labels, _vector


@typeguard.typechecked
class WriteJagged:
    def __init__(
        self: Self,
        path: str,
        name: str,
        dtype: DTypeLike | None = None,
        shape: tuple[int, ...] | None = None,
        check: str | None = None,
    ) -> None:
        self.vector = _vector.WriteVector(
            path,
            name,
            shape=shape,
            dtype=dtype,
            check=check,
        )
        self._index = _vector.WriteVector(
            path,
            _labels.get_index_name(name),
            shape=(),
            dtype=np.int64,
            check=check,
        )

    def __enter__(self: Self) -> Self:
        self.vector.__enter__()
        self._index.__enter__()
        self._index.append(np.array(0, dtype=np.int64))
        return self

    def __exit__(self: Self, *args) -> None:
        self.vector.__exit__(*args)
        self._index.__exit__(*args)

    def append(self, value: np.ndarray) -> None:
        # first axis is treated as variable length
        self.vector.extend(value)
        self._index.append(np.array(self.vector._index, dtype=np.int64))


@typeguard.typechecked
class ReadJagged:
    def __init__(self: Self, path: str, name: str) -> None:
        self.vector = _vector.read_vector(path, name)
        self._index = _vector.read_vector(path, _labels.get_index_name(name))

    @overload
    def __getitem__(self: Self, index: int) -> np.memmap:
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> list[np.memmap]:
        ...

    def __getitem__(
        self: Self,
        index: int | np.ndarray,
    ) -> np.memmap | list[np.memmap]:
        if isinstance(index, int):
            index = np.array([index])
            return self[index][0]

        assert isinstance(index, np.ndarray)

        if index.ndim == 0:
            error = "Index must be one dimensional."
            raise ValueError(error)

        if index.dtype == bool:
            index = np.where(index)[0]

        if np.any(index < 0):
            error = "Index must be positive."
            raise ValueError(error)

        if index.max() >= len(self._index) - 1:
            error = "Index out of bounds."
            raise IndexError(error)

        indicies = zip(self._index[index], self._index[index + 1])

        # preallocate result, this should be a bit faster
        result = [None] * len(index)

        for i, (s, e) in enumerate(indicies):
            result[i] = self.vector[s:e]  # type: ignore

        return result  # type: ignore

    def __len__(self) -> int:
        return len(self._index) - 1

    def __iter__(self) -> Any:
        for i in range(len(self)):
            yield self[i]


@typeguard.typechecked
class WriteShaped:
    def __init__(
        self: Self,
        path: str,
        name: str,
        dtype: DTypeLike | None = None,
        check: str | None = None,
    ) -> None:
        self.vector = WriteJagged(
            path,
            name,
            shape=(),
            dtype=dtype,
            check=check,
        )
        self._shape = WriteJagged(
            path,
            _labels.get_shape_name(name),
            shape=(),
            dtype=np.int64,
        )

    def __enter__(self: Self) -> Self:
        self.vector.__enter__()
        self._shape.__enter__()
        return self

    def __exit__(self: Self, *args) -> None:
        self.vector.__exit__(*args)
        self._shape.__exit__(*args)

    def append(self, value: np.ndarray) -> None:
        self.vector.append(value.reshape(-1))
        self._shape.append(np.array(value.shape, dtype=np.int64))


@typeguard.typechecked
class ReadShaped:
    def __init__(self: Self, path: str, name: str) -> None:
        self.vector = ReadJagged(path, name)
        self._shape = ReadJagged(path, _labels.get_shape_name(name))

    @overload
    def __getitem__(self: Self, index: int) -> np.memmap:
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> list[np.memmap]:
        ...

    def __getitem__(
        self: Self,
        index: int | np.ndarray,
    ) -> np.memmap | list[np.memmap]:
        if isinstance(index, int):
            index = np.array([index])
            return self[index][0]

        assert isinstance(index, np.ndarray)

        if index.ndim == 0:
            error = "Index must be one dimensional."
            raise ValueError(error)

        if index.dtype == bool:
            index = np.where(index)[0]

        if np.any(index < 0):
            error = "Index must be positive."
            raise ValueError(error)

        shape = self._shape[index]
        arrays = self.vector[index]

        for i, array in enumerate(arrays):
            arrays[i] = array.reshape(shape[i])  # type: ignore

        return arrays

    def __len__(self) -> int:
        return len(self.vector)

    def __iter__(self) -> Any:
        for i in range(len(self)):
            yield self[i]


@typeguard.typechecked
class WriteStructured:
    def __init__(
        self: Self,
        path: str,
        name: str,
        get_structure: Callable[[np.ndarray], np.ndarray],
        check: str | None = None,
    ) -> None:
        self.get_structure = get_structure
        self.data = WriteShaped(path, name, check=check)
        self.structure = WriteShaped(
            path,
            _labels.get_structured_name(name),
            check=check,
        )

    def __enter__(self: Self) -> Self:
        self.data.__enter__()
        self.structure.__enter__()
        return self

    def __exit__(self: Self, *args) -> None:
        self.data.__exit__(*args)
        self.structure.__exit__(*args)

    def append(self, value: np.ndarray) -> None:
        self.data.append(value)
        self.structure.append(self.get_structure(value))


@typeguard.typechecked
class ReadStructured:
    def __init__(
        self: Self,
        path: str,
        name: str,
        reconstruct: Callable[[np.ndarray, np.ndarray], Any],
    ) -> None:
        self.reconstruct = reconstruct
        self.data = ReadShaped(path, name)
        self.structure = ReadShaped(path, _labels.get_structured_name(name))

    @overload
    def __getitem__(self: Self, index: int) -> Any:
        ...

    @overload
    def __getitem__(self: Self, index: np.ndarray) -> list[Any]:
        ...

    def __getitem__(self: Self, index: int | np.ndarray) -> Any | list[Any]:
        if isinstance(index, int):
            index = np.array([index])
            return self[index][0]

        data = self.data[index]
        structure = self.structure[index]

        for i, (d, s) in enumerate(zip(data, structure)):
            data[i] = self.reconstruct(d, s)

        return data



import math
from typing import Any, Callable, Generator, Self

import numpy as np
import typeguard

from . import _jagged, _labels, _vector


@typeguard.typechecked
def load_memmaps(
    path: str,
    keys: set[str] | None = None,
) -> dict[str, np.memmap | _jagged.ReadJagged | _jagged.ReadShaped]:
    meta = _labels.safe_load(path)
    keys = keys or meta["arrays"].keys()
    data = {}

    for k in keys:
        if _labels.get_index_name(k) in meta["arrays"]:
            data[k] = _jagged.ReadJagged(path, k)
            continue

        if _labels.get_shape_name(k) in meta["arrays"]:
            data[k] = _jagged.ReadShaped(path, k)
            continue

        if _labels.get_structured_name(k) in meta["arrays"]:
            # Skipping structured data.
            # This is because we do not know how to reconstruct it here.
            continue

        data[k] = _vector.read_vector(path, k)

    return data


@typeguard.typechecked
def permutation_bands(length: int) -> Callable[[np.ndarray], np.ndarray]:
    bands_total = int(math.log2(length) ** 2)
    band_offsets = np.random.uniform(size=bands_total)
    band_offsets = (band_offsets * length).astype(np.int64)

    @typeguard.typechecked
    def permutation_func(indicies: np.ndarray) -> np.ndarray:
        indicies_offsets = indicies // bands_total
        indicies_remainders = indicies % bands_total

        indicies_offsets += band_offsets[indicies_remainders]
        indicies_offsets %= length // bands_total

        return indicies_offsets * bands_total + indicies_remainders

    return permutation_func


@typeguard.typechecked
def permutation_modulus(
    length: int,
    modulus: int = int(2**17 - 1),
) -> Callable[[np.ndarray], np.ndarray]:
    modulus = min(int(modulus), length)
    subpermutation = np.random.uniform(size=modulus).argsort()

    perm = np.random.uniform(size=modulus)
    perm = np.argsort(perm)

    @typeguard.typechecked
    def permute(indicies: np.ndarray) -> np.ndarray:
        indicies_offsets = indicies // modulus
        indicies_remainders = indicies % modulus

        indicies_remainders = subpermutation[indicies_remainders]

        result = indicies_offsets * modulus + indicies_remainders
        return result % length

    return permute


@typeguard.typechecked
def permutation_combination(length: int) -> Callable[[np.ndarray], np.ndarray]:
    perm1 = permutation_modulus(length)
    perm2 = permutation_bands(length)

    @typeguard.typechecked
    def perm(x: np.ndarray) -> np.ndarray:
        x = perm1(x)
        x = perm2(x)
        return x

    return perm


@typeguard.typechecked
def generate_batch_indicies(
    length: int,
    batch_size: int,
    drop_remainder: bool = True,
) -> Generator[np.ndarray, None, None]:
    for i in range(0, length - batch_size + 1, batch_size):
        yield np.arange(i, i + batch_size, dtype=np.int64)

    if not drop_remainder:
        remainder = length % batch_size
        yield np.arange(length - remainder, length, dtype=np.int64)


def _safe_take_set(s) -> Any:
    s = set(s)

    if len(s) != 1:
        raise ValueError("Set does not have size 1.")

    return next(iter(s))


@typeguard.typechecked
class SimpleLoader:
    def __init__(
        self: Self,
        data: dict,
        batch_size: int = 32,
    ) -> None:
        assert "_index" not in data, "Key '_index' is already used internally."
        self.data = data
        self.data_len: int = _safe_take_set(len(v) for v in self.data.values())
        self.batch_size = batch_size

    def __iter__(self: Self) -> Generator[dict[str, np.ndarray], None, None]:
        data_len_rounded = self.data_len - self.batch_size + 1

        for i in range(0, data_len_rounded, self.batch_size):
            indicies = np.arange(i, i + self.batch_size, dtype=np.int64)
            batch = {k: v[indicies] for k, v in self.data.items()}
            yield batch | {"_index": indicies}

        remainder = self.data_len % self.batch_size
        if remainder != 0:
            indicies = np.arange(
                self.data_len - remainder,
                self.data_len,
                dtype=np.int64,
            )
            batch = {k: v[indicies] for k, v in self.data.items()}
            yield batch | {"_index": indicies}

    def __len__(self: Self) -> int:
        result = self.data_len // self.batch_size

        if self.data_len % self.batch_size != 0:
            result += 1

        return result


class FixedSeed:
    def __init__(self, seed: int):
        self.seed = seed

    def __enter__(self):
        self.state = np.random.get_state()
        np.random.seed(self.seed)

    def __exit__(self, *_):
        np.random.set_state(self.state)


@typeguard.typechecked
class SplitLoader:
    def __init__(
        self,
        data,
        batch_size: int = 32,
        *,
        split: int = 0,
        fractions: tuple[float, ...] = (0.8, 0.1, 0.1),
        subindex: np.memmap | None = None,
        split_seed: int = 0,
        shuffle=True,
    ) -> None:
        self.data = data
        self.data_len: int = _safe_take_set(len(v) for v in self.data.values())
        self.batch_size = batch_size

        assert abs(sum(fractions) - 1.0) < 1e-6, "Fractions must sum to 1."
        self.split = split
        self.fractions = fractions
        self.subindex = subindex
        self.shuffle = shuffle

        if self.subindex is not None:
            self.data_len = len(self.subindex)

        with FixedSeed(split_seed):
            self.split_perm = permutation_combination(self.data_len)

        self.split_length = int(self.data_len * fractions[split])
        self.split_start = int(
            sum(self.data_len * f for f in fractions[:split]),
        )

    def __iter__(
        self: Self,
    ) -> Generator[dict[str, np.ndarray | np.memmap], None, None]:
        data_len_rounded = self.split_length - self.batch_size + 1

        if self.shuffle:
            perm_shuffle = permutation_combination(self.split_length)
        else:
            perm_shuffle = lambda x: x

        for i in range(0, data_len_rounded, self.batch_size):
            indicies = np.arange(i, i + self.batch_size, dtype=np.int64)
            indicies = self.split_perm(indicies + self.split_start)
            indicies = perm_shuffle(indicies)

            if self.subindex is not None:
                indicies = self.subindex[indicies]

            batch = {k: v[indicies] for k, v in self.data.items()}
            yield batch | {"_index": indicies}

    def __len__(self) -> int:
        return self.split_length // self.batch_size



try:
    import uproot
except ImportError:
    error = "'Uproot' is a requirement for 'memmpy.reroot', but was not found."
    raise ImportError(error)

import dataclasses
import hashlib
import os
from typing import Any

import awkward as ak
import numpy as np
import tqdm
import typeguard

from . import _labels, _loader, _subset, _vector

__all__ = [
    "RFileConfig",
    "load_root",
]


@dataclasses.dataclass(frozen=True)
class RFileConfig:
    path: str
    tree: str = dataclasses.field(default="nominal_Loose")
    metadata: dict[str, str] = dataclasses.field(default_factory=dict)


def encode_metadata(
    metadata: dict[str, str],
    size: int,
) -> dict[str, np.ndarray]:
    metadata_encoded = {}

    for k, v in metadata.items():
        hashed = _labels.hash_64bit(v.encode())
        metadata_encoded[k] = np.full(
            size,
            hashed,
            dtype="uint64",
        )

    return metadata_encoded


@typeguard.typechecked
def memmap_vectors_root(
    root_files: list[RFileConfig],
    path_mmap: str,
    keys: set[str],
    keys_padded: dict[str, tuple[int, Any]] | None = None,
    aliases: dict[str, str] | None = None,
    recreate: bool = False,
) -> None:
    keys_padded = keys_padded or {}
    aliases = aliases or {}
    keys = keys | set(keys_padded.keys())

    hasher = hashlib.sha256()
    hasher.update(str(aliases).encode())

    for rfile in root_files:
        hasher.update(os.path.basename(rfile.path).encode())
        hasher.update(str(os.path.getmtime(rfile.path)).encode())
        hasher.update(str(os.path.getsize(rfile.path)).encode())
        hasher.update(str(rfile.tree).encode())
        hasher.update(str(rfile.metadata).encode())

    hashed = hasher.hexdigest()

    if not recreate:
        meta = _labels.safe_load(path_mmap)

        for k in keys:
            if k not in meta["arrays"]:
                break

            if meta["arrays"][k]["check"] != hashed:
                break

            if k in keys_padded:
                if meta["arrays"][k]["shape"][-1] != keys_padded[k][0]:
                    break

        else:
            path_mmap = os.path.abspath(path_mmap)
            print(f"Using cached files in '{path_mmap}' for {list(keys)}.")
            return

    with _vector.WriteVectorDict(path_mmap, check=hashed) as vectors:
        for rfile in tqdm.tqdm(root_files):
            tqdm.tqdm.write(f"processing '{rfile.path}'")

            with uproot.open(rfile.path) as file:  # type: ignore
                data_root = file[rfile.tree].arrays(  # type: ignore
                    keys,
                    aliases=aliases,
                    library="ak",
                    how=dict,
                )
                data_numpy = {}

                for k, v in data_root.items():
                    if k in keys_padded:
                        assert v.ndim == 2
                        pad_size, pad_value = keys_padded[k]
                        array = ak.pad_none(v, pad_size, clip=True)
                        array = ak.fill_none(array, pad_value)
                        data_numpy[k] = array.to_numpy()  # type: ignore
                        continue

                    assert v.ndim == 1
                    data_numpy[k] = v.to_numpy()

                size = next(iter(data_numpy.values())).shape[0]

                data_path = encode_metadata(rfile.metadata, size)
                vectors.expand(data_numpy | data_path)

        print("Saving...")

    # add hahses
    metadata_values = set()

    for rfile in root_files:
        metadata_values |= set(rfile.metadata.values())

    for value in metadata_values:
        _labels.add_hash(path_mmap, value)

    print("Saving done.")


@typeguard.typechecked
def load_root(
    root_files: list[RFileConfig],
    path_mmap: str,
    keys: set[str],
    keys_padded: dict[str, tuple[int, Any]] | None = None,
    tcut: str | None = None,
    aliases: dict[str, str] | None = None,
    recreate: bool = False,
    batch_size: int = 32,
    split: int = 0,
    fractions: tuple[float, ...] = (0.8, 0.1, 0.1),
    split_seed: int = 0,
    shuffle=True,
    cut_constants: dict[str, Any] | None = None,
) -> _loader.SplitLoader:
    keys_mmemap = keys
    if tcut is not None:
        keys_mmemap = keys_mmemap | _subset.get_symbols(tcut)

    memmap_vectors_root(
        root_files,
        path_mmap,
        keys_mmemap,
        keys_padded,
        aliases,
        recreate,
    )
    keys_metadata = set()
    for rfile in root_files:
        keys_metadata |= set(rfile.metadata.keys())

    data = _loader.load_memmaps(path_mmap, keys | keys_metadata)

    subindex = None
    if tcut is not None:
        _subset.compute_cut_batched(path_mmap, tcut, constants=cut_constants)
        subindex = _vector.read_vector(path_mmap, tcut)

    return _loader.SplitLoader(
        data,
        batch_size,
        split=split,
        fractions=fractions,
        subindex=subindex,
        split_seed=split_seed,
        shuffle=shuffle,
    )



import ast
import hashlib
from typing import Any

import numpy as np
import tqdm
import typeguard

from . import _labels, _loader, _vector


@typeguard.typechecked
def get_symbols(expression: str) -> set[str]:
    tree = ast.parse(expression, mode="eval")
    walk = ast.walk(tree)

    symbols = set()
    for node in walk:
        if isinstance(node, ast.Name):
            symbols.add(node.id)

    return symbols


@typeguard.typechecked
def compute_cut_batched(
    path: str,
    expression: str,
    *,
    batch_size: int = 1024 * 128,
    write_type=_vector.WriteVector,
    constants: dict[str, Any] | None = None,
    recreate: bool = False,
) -> np.memmap:
    constants = constants or {}

    keys = get_symbols(expression)
    keys = keys - set(constants.keys())

    hasher = hashlib.sha256()
    for key in sorted(keys):
        hasher.update(key.encode())

    hashed = hasher.hexdigest()

    if not recreate:
        meta = _labels.safe_load(path)

        if expression in meta["arrays"]:
            if meta["arrays"][expression]["check"] == hashed:
                print(f"Using cached cut: {expression}.")
                return _vector.read_vector(path, expression)

    print(f"Computing cut: {expression}.")
    with write_type(path=path, name=expression, check=hashed) as writer:
        batch_loader = _loader.SimpleLoader(
            _loader.load_memmaps(path, keys),
            batch_size,
        )

        for batch in tqdm.tqdm(batch_loader):
            result = eval(expression, batch | constants, {})
            index = batch["_index"][result]
            writer.extend(index)

    return _vector.read_vector(path, expression)



import math
import os
import tempfile
from typing import Self

import numpy as np
import typeguard
from numpy.typing import DTypeLike

from . import _labels


@typeguard.typechecked
class WriteVector:
    def __init__(
        self,
        path: str,
        name: str,
        shape: tuple[int, ...] | None = None,
        dtype: DTypeLike | None = None,
        check: str | None = None,
    ) -> None:
        self.path = path
        self.name = name
        self.shape = shape
        self.dtype = dtype
        self.check = check
        self.enter = False
        self.fixed = False

    def fix_shape_and_dtype(self, shape: tuple[int, ...], dtype: DTypeLike):
        if self.fixed:
            return

        self.shape = self.shape or shape
        self.dtype = self.dtype or dtype

        self._file = tempfile.NamedTemporaryFile()
        self._mmap = np.memmap(
            self._file,
            self.dtype,
            "w+",
            shape=(1, *self.shape),
        )

        self._index = 0
        self.fixed = True

    def __enter__(self: Self) -> Self:
        self.enter = True
        return self

    def extend(self: Self, value: np.ndarray) -> Self:
        self.fix_shape_and_dtype(value.shape[1:], value.dtype)
        assert self.enter, "WriteVector should be used as context manager."

        if value.dtype != self._mmap.dtype:
            error = f"Got dtype {value.dtype} != {self._mmap.dtype}."
            raise ValueError(error)

        if value.shape[1:] != self._mmap.shape[1:]:
            error = f"Got shape {value.shape[1:]} != {self._mmap.shape[1:]}."
            raise ValueError(error)

        length = self._index + value.shape[0]
        length = 2 ** math.ceil(math.log2(length))

        self._file = tempfile.NamedTemporaryFile()
        mmap = np.memmap(
            self._file,
            self.dtype,
            "w+",
            shape=(length, *self._mmap.shape[1:]),
        )

        mmap[: self._index] = self._mmap[: self._index]
        self._mmap = mmap

        self._mmap[self._index : self._index + value.shape[0]] = value
        self._mmap.flush()

        self._index += value.shape[0]
        return self

    def append(self: Self, value: np.ndarray) -> Self:
        return self.extend(value[None])

    def __exit__(self: Self, exc_type, *_) -> None:
        if exc_type is not None:
            return

        mmap_path = _labels.get_array_path(self.path, self.name)

        os.makedirs(self.path, exist_ok=True)
        mmap = np.memmap(
            mmap_path,
            self.dtype,
            "w+",
            shape=(self._index, *self._mmap.shape[1:]),
        )
        mmap[:] = self._mmap[: self._index]
        mmap.flush()

        _labels.save_array_metadata(
            self.path,
            self.name,
            shape=mmap.shape,
            dtype=mmap.dtype,
            check=self.check or _labels.get_timestamp(),
        )


@typeguard.typechecked
class WriteVectorDict:
    def __init__(self: Self, path: str, check=None):
        self.path = path
        self.check = check
        self.fixed = False

    def fix_keys(self, keys):
        if self.fixed:
            return

        self.vectors = {}
        for k in keys:
            self.vectors[k] = WriteVector(self.path, k, check=self.check)
            self.vectors[k].__enter__()

        self.fixed = True

    def __enter__(self: Self) -> Self:
        self.enter = True
        return self

    def expand(self: Self, data: dict[str, np.ndarray]) -> Self:
        if not self.fixed:
            self.fix_keys(data.keys())

        assert set(data.keys()) == set(self.vectors.keys())
        assert len(set(v.shape[0] for v in data.values())) == 1

        for key, value in data.items():
            self.vectors[key].extend(value)

        return self

    def __exit__(self: Self, *args) -> None:
        for vector in self.vectors.values():
            vector.__exit__(*args)


@typeguard.typechecked
def read_vector(path: str, name: str) -> np.memmap:
    mmap_path, dtype, shape = _labels.read_array_metadata(path, name)
    return np.memmap(mmap_path, dtype, "r", shape=shape)



from ._jagged import ReadJagged, ReadShaped, WriteJagged, WriteShaped
from ._loader import SimpleLoader, SplitLoader, load_memmaps
from ._subset import compute_cut_batched
from ._vector import WriteVector, read_vector

__all__ = [
    "WriteVector",
    "read_vector",
    "WriteJagged",
    "ReadJagged",
    "WriteShaped",
    "ReadShaped",
    "SimpleLoader",
    "SplitLoader",
    "load_memmaps",
    "compute_cut_batched",
]

__version__ = "0.1.2"
